#!/usr/bin/env python
''' 
 Usage: Instantiate the Matrix class, either with:
 1. The name of the selected package or
 2. Without any parametrs and let the initialisation find an installed matrix
    package.
         
 The following packages are suppoted: "numarray", "Numeric" and "numpy".
 Class to pass essential matrix operation functions to the calling program
 using either of:
 1. numarray module
 2. Numeric and LinearAlgebra modules. Please note that in ubuntu, which is
    Debian based, the LinearAlgebra module is installed when the
    python-numeric-ext and python-numeric-ext-dbg are installed.
 3. numpy (with  numpy.linalg)
The supplied functions are:
 1. A = zeroM(m, n)    = create a m by n matrix A of zeros
 2. B = transposeM(A)  = transpose matrix A and store in B
 3  C = multM(A, B)    = multiply matrix A by matrix B and store in C
 4. B = inverseM(A)    = invert matrix A and store in B
 5. u = solveM(A, Rhs) = solve linear equations. Eqn matrix = A; rhs = Rhs; u = solution
 6. PrintM(A)          = neatly print matrix A; Optional second parameter for line length.
 
 The class is flexible - it can be instantiated either with matrix module specified
 by a parameter, or it can be called without parameters. In the latter circumstance,
 the instantiation of the class Matrix will search for the first available installed
 matrix operations package.
 Thanks to this wrapper program, the user need not study the calling procedures for
 various matrix operations - all functions, regardless of the matrix package 
 conventions, are called in exactly the same format. This is the principal gain
 as it oviates the need to study the various notations of different matrix packages.
'''
#matrixM2.py
#
# Copyright (C) 2008 Algis Kabaila
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# $Id: matrixM.py 168 2009-01-27 09:39:52Z kabalg $
#
# matrixM.py - a module of a subset of matrix functions in a general
# format which are the same regardless of the module used.
# Working directory is ~Documents/akabaila/utils/trunk
# SVN repository http://code.proams.org/proams/akabaila/utils/trunk

import sys

class Matrix:
    '''
A wrapper class to give the user uniform syntax for matrix operations in one of 
matrix the following matrix packages: [numarray, Numeric, numpy] 
    ''' 
    def __init__(self, qprint, flag = ''):
        '''
An available matrix package is selected. Optionally it can be specified as a parameter, 
one of ['numpy', 'Numeric', 'numarray'] during instantiation of the class.
        '''
        self.qprint = qprint
        keywordList = ['numpy', 'Numeric', 'numarray']
        if flag == '': # then matrix package name is not specified- auto determine:
            ok = False
            if not ok:
                try:
                    import numpy as _num
                    import numpy.linalg as _LA
                    ok = True; flag = 'numpy'
                except ImportError:
                    pass
            if not ok:
                try:
                    import Numeric as _num
                    import LinearAlgebra as _LA
                    ok = True; flag = 'Numeric'
                except ImportError:
                    pass
            if not ok:
                try:
                    import numarray as _num
                    import numarray.linear_algebra as _LA
                    ok = True; flag = 'numarray'
                except ImportError:
                    # ok is False. No package could be imported, flag == ''.
                    assert not ok
                    assert flag == ''
                    print 'No package could be imported.'
                    print 'Tried numpy, Numeric and numarray'
                    print 'Most likely none fully installed'
                    print 'Program terminates now.'
                    sys.exit(['No package available.'])                  

        else:  # flag has been specified by the calling program. Is it valid?
            if not (flag in keywordList):
                print 'You specified package ', flag, ' which is not supported'
                print 'by this "Matrix" class. The program now terminates.'
                sys.exit(['Unsupported package specified.'])
            assert (flag in keywordList)

            if flag == 'numpy':
                try:
                    import numpy as _num
                    import numpy.linalg as _LA                    
                except ImportError:
                    print 'Program terminates now.'
                    print 'numpy or numpy.linalg could not be imported.'
                    sys.exit(['Import of numpy failed.'])

            if flag == 'Numeric':
                try:                             
                    import Numeric as _num
                    import LinearAlgebra as _LA
                except ImportError:       
                    print 'Numeric or LinearAlgebra could not be imported.'
                    print 'Program terminates now.'                    
                    sys.exit(['Import of Numeric failed'])

            if flag == 'numarray':
                try:
                    import numarray as _num
                    import numarray.linear_algebra as _LA
                except ImportError:
                    print 'numarray or numarray.linear_algebra could not be imported.'
                    print 'Program terminates now.'
                    sys.exit(['Import of numarray failed'])
                    
        self._num = _num
        self._LA  = _LA
        self.flag = flag
            
# General methods-begin ----------------------------------
#    def printM(self, M, LineLen=5):  # Notice the default value, which is changeable
#        '''
#printM(A matrix, [Line length]) -> Neatly prints matrix M of size (m x n). Line length can be 
#specified as second parameter.   
#        '''
#        m,n = len(M), len(M[0])
#        print 'A matrix of dimensions (m x n), where m, n, LineLen = ', m, n, LineLen
#        for i in range(0,m):
#            print
#            for j in range(0,n):
#                print ' %15.5E ' % M[i][j],
#                if ( (j+1) % LineLen) == 0: print # New line
#            if (j==(n-1)) and ((j+1) > LineLen) : print
#        print;print

#    def filePrintM(self, M, f, LineLen=5):        
#        '''
#filePrintM(A matrix, [Line length]) -> Neatly file-prints matrix M of size (m x n).
#Line length may be specified as second parameter (default is 5).   
#        '''
#        m,n = len(M), len(M[0])
#        t = 'A matrix of dimensions (m x n), where m, n, LineLen =  %i %i %i ' % (m, n, LineLen)
#        f.write(t + '\n')
#        for i in range(0,m):
#            f.write('\n')
#            for j in range(0,n):
#                f.write(' %15.5E ' % M[i][j])
#                if ( (j+1) % LineLen) == 0: f.write('\n') # New line
#            if (j==(n-1)) and ((j+1) > LineLen) : f.write('\n')
#        f.write('\n'); f.write('\n')


    def zerosM(self, m, n):
        '''
(m,n) --> (mxn) matrix, filled with floating zeros. 
        '''
        _num = self._num; flag = self.flag
        if flag == "numarray":
            return _num.zeros((m,n),_num.Float)
        elif flag == "Numeric":
            return _num.zeros((m,n),typecode="d")
        elif flag == "numpy":
            return _num.zeros((m,n),dtype=float)
        else:
            self.wrongChoice()
          
    def transposeM(self,A):
        '''
(A) matrix -- transpose of matrix A 
        '''
        return self._num.transpose(A)

    def multM(self,A,B):
        '''
(A,B) mlultiplication compatible matrices --> [A x B] matrix product.   
        '''
        flag = self.flag; _num = self._num
        if flag == "numarray":
            return _num.matrixmultiply(A,B)
        elif flag == "Numeric":
            return _num.dot(A,B)
        elif flag == "numpy":
            return _num.dot(A,B)
        else:
            self.wrongChoice()

    def inverseM(self, A):
        '''
(A) non singular square matrix A --> inverse of A.  
        '''
        flag = self.flag; _LA = self._LA
        if flag == 'numarray' or flag == 'Numeric':
            return _LA.inverse(A)
        elif flag == "numpy":
            return _LA.inv(A)
        else:
            self.wrongChoice()

    def solveM(self,A,Rhs):
        '''
(A matrix of coefficients,Rhs right hand sides of equation [A]x[X] = [Rhs] --> [X]  
        '''
        flag = self.flag; _LA = self._LA
        if flag == "numarray":
            return _LA.solve_linear_equations(A,Rhs)
        elif flag == "Numeric":
            return _LA.solve_linear_equations(A,Rhs)
        elif flag == "numpy":
            return _LA.solve(A,Rhs)
        else:
            self.wrongChoice()
    
    def wrongChoice(self):
        '''
wrongChoice() prints a message and exits the program if non-existent matrix module is chosen.
        '''
        print"Invalid parameter specified for setup. Parameter = ", self.flag
        sys.exit(['invalid parameter'])
    
#    def printer_m(self, fo, qprint, M, LineLen=5):        
#        '''
#printer_m(fo, A matrix, [Line length]) -> Neatly file-prints matrix M of size (m x n).
#Line length may be specified as second parameter (default is 5).   
#        '''
#        m, n = len(M), len(M[0])
#        t = 'A matrix of dimensions (m x n), where m, n, LineLen =  %i %i %i ' % (m, n, LineLen)
#        fo.write(t + '\n')
#        qprint(t + '\n')
#        for i in range(0,m):
#            fo.write('\n')
#            qprint('\n')
#            sumj = ''
#            for j in range(0,n):
#                fo.write(' %12.5E ' % M[i][j])
#                sumj += ' %12.5E ' % (M[i][j])
#            qprint(sumj)
#            if ( (j+1) % LineLen) == 0: 
#                fo.write('\n') # New line
#                qprint('\n') # New line
#            if (j==(n-1)) and ((j+1) > LineLen) : 
#                fo.write('\n')
#                qprint('\n')
#        fo.write('\n\n')
#        qprint('\n')

    def printm(self, fo, qprint, M, LineLen=5):        
        '''
printm(fo, A matrix, [Line length]) -> Neatly file-prints matrix M of size (m x n).
Line length may be specified as second parameter (default is 5).   
        '''
        m, n = len(M), len(M[0])
        t = 'A matrix of dimensions (m x n), where m, n, LineLen =  %i %i %i ' % (m, n, LineLen)
        fo.write(t + '\n')
        qprint(' ')
        for i in range(m):
            fo.write('\n')
            line = ''
            for j in range(n):
                fo.write(' %12.5E ' % M[i][j])
                line += ' %12.5E ' % (M[i][j])
                if ( (j+1) % LineLen) == 0 or ((j + 1) == n): 
                    fo.write('\n') # New line
                    qprint(line)
                    if j < (n - 1):
                        line = ''
                if (j==(n - 1)) and ((j + 1) > LineLen) : 
                    fo.write('\n')
#                    qprint(line)
                    qprint(' ')
        fo.write('\n')
        
# General methods-end    ----------------------------------

if __name__ == "__main__":
    '''
Some testing examples if the module is run as '__main__' program
# The same program testin installed in the style of  tester.py
    '''
#    global result
#    result = ''

#    fs = sys.stdout   
#
#    def qprint(line):
#        global result
#        result += (line + '\n')
#    fs.write(line + '\n')
#    _m =  Matrix(qprint, 'numpy')
#    qprint ('----------------------------------------------------------------\n')
#    qprint('3x2 matrix of zeros, a\n')
#    a = _m.zerosM(3,2)
#    _m.printer_m(fo, qprint, a)
#    qprint ('Transpose of a\n')
#    at = _m.transposeM(a) 
#    _m.printer_m(fo, qprint, at)
##    def printer_m(self, fo, qprint, M, LineLen=5):     
#    A = _m.zerosM(3,3)
#    # Fill A with numbers
#    A[0,0] = 0.0;    A[0,1] = 1.0;    A[0,2] = 2.0
#    A[1,0] = 12.0;    A[1,1] = 6.0;    A[1, 2]= 0.0
#    A[2, 0] = 0.0;  A[2, 1] = 0.0; A[2,2] = 7.0
#    qprint('----------------------------------------------------------------\n')
#    qprint ('A non singular matrix A.\n')
#    _m.printer_m(fo, qprint, A)
#    qprint ('----------------------------------------------------------------\n')    
#    B = _m.inverseM(A)    
#    qprint ('[B] = inverse([A]).\n')
#    _m.printer_m(fo, qprint, B)
#    qprint ('----------------------------------------------------------------\n')
#    qprint ('  I = matrixmultiply(A,B) = Unit matrix.\n')
#    I = _m.multM(A,B)
#    _m.printer_m(fo, qprint, I)
#    qprint ('----------------------------------------------------------------\n')    
#    qprint('Equation solving, using the inverse.\n')
#    Rhs = _m.zerosM(3,1)
#    Rhs[0] = 8.; Rhs[1] = 24. ;Rhs[2] = 21
#    qprint ('Right hand side vector Rhs.\n')4
#    qprint (Rhs.__str__()); qprint('\n')
#    qprint ('----------------------------------------------------------------\n')        
#    v = _m.multM(B,Rhs)
#    qprint ('The solution = [inverse of A] x [Rhs]\n')
#    qprint(v.__str__())
#    qprint ('\n----------------------------------------------------------------\n')
#    qprint ('Solution of simultaneous equations without matrix inversion.\n')
#    vv = _m.solveM(A,Rhs)
#    qprint(vv.__str__())
#    qprint ('\n----------------------------------------------------------------\n')
#    qprint ('(3 x 2) matrix [B] with some non zero values.\n')
#    B = _m.zerosM(3,2)
#    x = 0.
#    for i in range(0,3):
#      for j in range(0,2):
#          x = x + 1.
#          B[i,j] = x
#    _m.printer_m(fo, qprint, B)
#    qprint ('----------------------------------------------------------------\n')
#    qprint ('[B] = transpose([B]).\n')
#    B = _m.transposeM(B)
#    _m.printer_m(fo, qprint, B)
#    qprint ('----------------------------------------------------------------\n')
## Add tests for printer_m(self, fo, M, LineLen=5)
#    header = '[B] printed by printer_m(fo, B)\n'
#    qprint(header)
#    fo.write('\n' + header + '\n\n')
#    _m.printer_m(fo, qprint,  B)
#    fo.close()
#    qprint ('----------------------------------------------------------------\n')    
#    print '----------------------------------------------------------------'
#    print 
#    print ('**********************************************************')
#    print("Testing routines for functions created")
#    print ('**********************************************************')
#    Try to print a 3x8 matrix (of course, after first generating it
#    a = _m.zerosM(3, 8)

# Test the printm method.

    fo = open('/dat/work/fem2/dummy.out', 'w')    
    
    def qprint(line):
        print line
        
    _m = Matrix(qprint, 'numpy')    
    a = _m.zerosM(3, 14)
    for i in range(3):
        for j in range(14):
            a[i, j] = i * 10.0 + j * 0.1
    _m.printm(fo, qprint, a)
    print('All done and finished.')
    fo.write('All done and finished.\n')
#    print result
    fo.close()
