#!/usr/bin/env python3
'''
joint_mod.py - module for trusses and frames as well as frames joined with 
trusses; trusses3d; potentially other alements. Used in program fem.py.
Nodal list processing changed as follows:
1. Read nodes into memory as a directory of lines.
2. Process each line and change every line in succession.
3. Calculate freedoms and store with the original nodes directory.

'''

# 
#
#        (c) Copyright 2006, 2008, 2009 Algis Kabaila.  All rights reserved.
#   
# This document is free; you can redistribute it and/or modify it under the terms
# of the GNU General Public License as published by the Free Software Foundation;
# either version 2 of the License, or (at your option) any later version.
#
# This document is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this document; if not, write to the Free Software Foundation, Inc.,
# 675 Mass Ave, Cambridge, MA 02139, USA.

import sys
#from core_mods import core_mod3 as cm_ 
from . import core_mod3 as cm_
from . import Matalg as _m
#from matrixM2 import Matrix

# -------------------------------------------------------------------
#DEBUG = True

#def d_print(val, ):
#    if DEBUG print 'dic* ', (val)D
#    if DEBUG print 'dic* ', (val)D
    
class My_exception(Exception):
    pass

class Material_list:
    '''Dictionary storage of num and E values: dict_e = {num: E}'''
    def __init__(self, glob):
        '''This is currently for one material property only - E!'''
        self.glob = glob
        self.dic = {}  # dict_e format: {num: E}
        
    def read_material_list(self):
        '''Read Youngs modulus and other properties, if any.'''
        fin  = self.glob.fin
        fout = self.glob.fout
        printline = self.glob.qprint
        for i in range(self.glob.no_materials):
            lst = cm_.checkList(fin, fout, printline, 'g')
            # checkList removes item 0 ('g')
            lss  = cm_.listProcess(fout, printline, lst, 'sif')
            if lss[0] in ('t', 'u', 'b'):
                self.dic[lss[1]] = lss[2]
                # added is one type tag lss[0]
            else:                
                raise My_exception('Unknown element type') 

class Section_list:
    '''Sequential list of section properties. Number represents the type.'''    
    def __init__(self, glob):
        self.glob = glob
        self.dic = {} # dic format: {num: (item1, item2, )
              
    def read_section_list(self):
        '''Read sectionList -> sectionList'''
        # 2009-11-01 ak stopped here.  #STOP#
        for i in range(self.glob.no_sections):
            fin  = self.glob.fin
            fout = self.glob.fout
            printline = self.glob.qprint
            lstt = cm_.checkList(fin, fout, printline, 'g')
#           lstt = ["tag", "no", "area", "M of I when applicable"]
            if lstt[0] in ('t', 'u'):
                lst = cm_.listProcess(fout, printline, lstt, 'sif')
                self.dic[lst[1]] = lst[2]
            elif lstt[0] in ('b'):
                lst = cm_.listProcess(fout, printline, lstt, 'siff')                
                self.dic[lst[1]] = (lst[2], lst[3])

# -------------------------------------------------------------------

class Node_list:
    '''Store members in a list to enable member insertion after initial calculation.'''
    def __init__(self, glob):
        self.nn = glob.no_nodes
        self.dic = {}
        self.freedom_count = 0
        self.properties = {}
        # directory format: {nodeNo : [bt, ms, x, y, flagX='f', flagY='f']}
        # Flag 'f' -> unrestrained freedom; 'r' -> restrained freedom.
        # Currently for truss pin joints (where free means freedom num):
        # {nodeNo: [bt, ms, x, y, freeX, freeY]}
        # for frame beams:
        # {nodeNo: [bt, ms, x, y, freeX, freeY, freeXY]}
        self.glob = glob
#        Chain onto read_node
        self.read_node_list()
        return
        
    def read_node_list(self):
        '''Read node_list -> node_list'''
        fin = self.glob.fin
        fout = self.glob.fout
        printline = self.glob.qprint
        cm_.getTitle(fin, fout, printline, 'nodes')
        count = 0
        while count < self.nn:
            line = fin.readline()[:-1] # Slice off the LF char '\n'
            cm_.printer(fout, printline,  line)
            if not ('#' in line):
                count += 1
                self.dic[count] = line.split()
# All nodes in chars stored in node_list.dic dictionary. 
# Pass the control passed to find_freedoms() method. 
        self.find_freedoms()
        return
        
    def main_parameters(self):
        instances_keys = list(self.glob.instances.keys())
#        print 'd* instances keys = ', instances_keys
# Often instances_keys has one value, viz for trusses keys = ['truss']        
        const = 3
        nodes_constants = {}
        for key in instances_keys:
            dic = self.glob.instances[key].dic
            tag = dic['tag']
            start_at = const + dic['coord_no']
            
# starts_at points at the first freedom. Here, dic is part of *instances*.
            nodes_constants[tag] = [start_at, dic['freedom_no'], dic['coord_no']]
#        return directory nodes_constants where directory[tag] = [start_at, freedom_nos]
        return nodes_constants
            
    def find_freedoms(self):
        '''Determine nodal freedoms, hence the number of structure
        freedoms. Also cast coordinates to float.'''
#        print '*** in find_freedoms'
        nodes_constants = self.main_parameters()
#        print 'd* nodes_constants = ', nodes_constants
#        print 'd* node list directory = ', self.dic
        keys = list(self.dic.keys())
#        print 'd* keys = ', keys
        keys.sort()
        freedom_count = 0
        for key in keys:           
            lst = self.dic[key]
#            print 'd* lst, raw =       ', lst 
            tag = lst[0]
            try:
                start_at = nodes_constants[tag][0]
                no_of_freedoms = nodes_constants[tag][1]
            except:
        #        print 'tag in mode list = ', tag, ' is probably wrong.'
                raise My_exception('wrong tag')
            freedoms = lst[start_at : start_at + no_of_freedoms]
#            print 'd* start_at, freedoms (raw) = ', start_at, ', ', freedoms
            address = 0
            for var in freedoms:
                if var == 'f':
                    freedom_count += 1
                    value = freedom_count
                elif var == 'r':
                    value = 0
#                Add here the slave node handling.
                elif 's' == lst[1]:
#                    print 'd** slave node freedoms lst = ', lst
                    try:
                        node_no = int(var)
                        abs_node_no = abs(node_no)
                    except:
                        print('echo lst = ', lst)                        
                        raise My_exception('Here freedom can be "f", "r" or integer.')
                    assert key > abs_node_no,\
                    'A node can be slave only to a processed node!'
                    master_lst = self.dic[abs_node_no]
                    temp = master_lst[start_at + address]
                    master_freedom = temp * cm_.sgn(node_no)
                    value = master_freedom
                else:
#                    printline(lst.__str__())
                    raise My_exception('invalid input line.')
#                print 'd* freedoms', freedoms
                freedoms[address] = value
# At the above point (start_at + address) points at the freedom being 
# processed in lst (- self.dic[key]). Here key is the node no.
                address += 1

                lst[start_at : ] = freedoms
            # convert coordinates to floats
            for i in range(3, (3 + nodes_constants[tag][2])):
                lst[i] = float(lst[i])
#            print 'd* lst, with freedom nos = ', lst
            
            self.dic[key] = lst
        self.glob.no_freedoms = freedom_count
#        print 'd* freedom number = ', freedom_count
        return

    def print_node_list(self):
        '''Print details of node list'''
#        fin  = self.glob.fin
#        fout = self.glob.fout
        printline = self.glob.qprint
        printline('\nNode list:\n')
        node_list = list(self.dic.keys())
        node_list.sort()
        for var in node_list:
            printline (self.dic[var].__str__() + '\n')

# -------------------------------------------------------------------

class Member_list:
    '''Store members in a list to enable member insertion after initial calculation.'''
    def __init__(self, glob):
        self.glob = glob
        self.node_list_dic = glob.node_list_dic
        self.dic = {} # {memberNo: [bt, endPt1, endPt2, noMaterial, noSection]}
# The above is supplemented during processing: tuple (d_x, d_y, [locationVector])
# is added to the dic[memberNo] value for 2d trusses and frames. For space trusses
# the tuple is (d_x, d_y, d_z, [locationVector])
#        self.props = {}
# member_list.props will need to be passed to assembler and solver.

    def read_member_list(self):
        '''Read memberList'''
        glob = self.glob
        fin  = glob.fin
        fout = glob.fout
        printline = glob.qprint
        cm_.getTitle(fin, fout, printline, 'members')
        for i in range(glob.no_members):
            lst = cm_.listRead(fin, fout, printline)
            lsx = cm_.listProcess(fout, printline, lst, 'siiiii')
            num = lsx[1]
            del(lsx[1])
            self.dic[num] = lsx

    def print_member_list(self):
        glob = self.glob
        fout  = glob.fout
        printline = glob.qprint
        member_list = list(self.dic.keys())
        member_list.sort()
        cm_.printer(fout, printline, 'Member list :')
        cm_.printer(fout, printline, 'memberNo : [bt, end1, end2, noMaterial, noSection]')
        for var in member_list:
            cm_.printer(fout, printline, str(var) + ':' + self.dic[var].__str__())
        
    def do_location_vectors(self):
        '''Determine locationVector for all members as well as d_x and d_y (and d_z for 3d members.'''
#        node_list_dic = self.node_list_dic
#        node_list_dic = self.node_list_dic
        member_list = list(self.dic.keys())
        member_list.sort()
        glob = self.glob
        props = {}  # local variable.
#        print 'd* member_list', member_list
#        print 'd* self.dic\n = ', self.dic
#        print 'd_x, d_y, e_mod, area, loc_vect - table'
        for i in member_list:
            member = self.dic[i]
#            print 'd* member = ', member
            if not (member[0] in ['t', 'b', 'u']):
                print('not a known member type in do_location_vectors')
                raise My_exception('Unknown element type encountered in member list.')
            inst = glob.instances
            if member[0] == 't':
                d_x, d_y, e_mod, area, loc_vect =   inst['truss'].find_element_properties\
                (member, glob.node_list_dic, glob.material_list_dic, glob.section_list_dic)
                props[i] = ['t', i, d_x, d_y, e_mod, area, loc_vect]
#                print ' %4i %12.4E %12.4E %12.4E %12.4E' % (i, d_x, d_y, e_mod, area), 
#                print 'Location vector = ', loc_vect
                prop = ' %4i %12.4E %12.4E %12.4E %12.4E LVec = %s\n' %\
                (i, d_x, d_y, e_mod, area, loc_vect.__str__())
                glob.qprint(prop)
                
            elif member[0] == 'b':
                d_x, d_y, e_mod, area, i_x, loc_vect = inst['beam'].find_element_properties\
                (member, glob.node_list_dic, glob.material_list_dic, glob.section_list_dic)
                props[i] = ['b', i, d_x, d_y, e_mod, area, i_x, loc_vect]
                glob.qprint (' %4i %12.4E %12.4E %12.4E %12.4E %12.4E' % (i, d_x, d_y, e_mod, area, i_x) + loc_vect.__str__())
#                sys.exit(['temporary program stop.'])
                
#                print 'not implemented.'
#                pass
            elif member[0] == 'u':
                truss3dim = inst['truss3dim']
                d_x, d_y, d_z, e_mod, area, loc_vect = truss3dim.find_element_properties\
                (member, glob.node_list_dic, glob.material_list_dic, glob.section_list_dic)                
                props[i] = ['u', i, d_x, d_y, d_z, e_mod, area, loc_vect]
                print(' %4i %12.4E %12.4E %12.4E %12.4E %12.4E' % (i, d_x, d_y, d_z, e_mod, area), end=' ') 
                print('Location vector = ', loc_vect)

#        print 'd* main loop completed.'    
#        print 'd* props ='
        return props
        
class Load_list:
    '''Methods to read and store loads.'''
    def __init__(self, glob, ii, node_list):   # nn):
        '''(nn = number of loaded nodes)'''
        self.ii = ii
        self.glob = glob
        self.node_list = node_list
        self.no_loads  = glob.no_loads[ii]    #nn
        self.dic = {} 

    def read_load_list(self):
        '''Read list of node loads: tag, nodeNo, X force, Y force, couple.'''
        fin  = self.glob.fin
        fout = self.glob.fout
        printline = self.glob.qprint
        dic = self.dic
        for i in range(self.no_loads):       
            lst = cm_.listRead(fin, fout, printline)
            # test cm_.printer
            cm_.printer(fout, printline, lst.__str__())
            # Discrimanate here between t and b for FEM
            if lst[0] == 't':
                if len(lst) > 4:
                    cm_.printer(fout, \
                        'WARNING: Truss nodes must not be subjected to couples.')
                    cm_.printer(fout, lst.__str__())
                    sys.exit(['too much data'])                    
                lsx = cm_.listProcess(fout, printline, lst, 'siff')
                num = lsx[1]
                nodalForces = lsx[2:]
                # nodalForces = [x, y, fX, fY]
                tmp = self.node_list.dic[num]
                node = tmp[-2:]
                # node = [freeX, freeY]
                # directory dic format for nodes:
                # {nodeNo : [bt, ms, x, y, flagX='f', flagY='f', flagXY]}
                # in the current #1 stage {nodeNo: [bt, ms, x, y, freeX, freeY, freeXY]}
                # (free = freedom) flagXY and stage #1 forceXY only exists in beam nodes.
            elif lst[0] in ['b', 'u']:
                lsx = cm_.listProcess(fout, printline, lst, 'sifff')
                # lsx = [tag, nodeNo, forceX, forceY, forceXY] for 'b' type element
                # lsx = [tag, nodeNo, forceX, forceY, forceZ]  for 'u' type element
                num = lsx[1]
                nodalForces = lsx[2:]
#                print '*dic* nodalForces = ', nodalForces
                # nodalForces = [forceX, forceY, forceXY]
                tmp = self.node_list.dic[num]
                node = tmp[-3:]
                # node = [freeX, freeY, freeXY]
            for i in range(len(nodalForces)):
                var = nodalForces[i]
                freedom = node[i]
                if var is not 0.0:
                    keys = list(dic.keys())
                    if abs(freedom) in keys:
                        dic[abs(freedom)] += cm_.sgn(freedom) * var
                    else:
                        dic[abs(freedom)] = cm_.sgn(freedom) * var           
        return dic
    
class Load_cases_list:
    '''Record all Load Cases.'''
    def __init__(self, glob, node_list):
        self.glob = glob
        self.node_list = node_list
        self.no_load_cases = glob.no_load_cases
        self.load_list_dicts = [] # list of LoadLists.dicts
        # loadLists.dicts format = {freedomNo: force}

    def do_load_matrix(self):
        "Create a load matrix."
        fout = self.glob.fout
        glob = self.glob
        printline = glob.qprint
        no_freedoms = glob.no_freedoms   
        printline ('Earlier calculated no_freedoms  ' + str(no_freedoms))  
#        loadMatrix = Matrix(glob.qprint).zerosM(no_freedoms, glob.no_load_cases)
        loadMatrix = _m.Matrix(no_freedoms, glob.no_load_cases)
        for ii in range(len(self.load_list_dicts)):
            dic = self.load_list_dicts[ii]
            key_list = list(dic.keys())
            for j in key_list:
                loadMatrix[j - 1, ii] = dic[j]
        return  loadMatrix

    def read_load_case(self):
        "Read load list for each Load Case."
        cm_.getTitle(self.glob.fin, self.glob.fout, self.glob.qprint, 'loads')
        for ii in range(self.no_load_cases):
            load_list = Load_list(self.glob, ii, self.node_list)  #.no_loads[ii])
            dic = load_list.read_load_list()  # read_load_list must do all the work
            self.load_list_dicts.append(dic)

# -------------------------------------------------------------------

class Prescribes:
    "Prescribed non-zero displacements for nodes."
    def __init__(self, glob, node_list):
        self.glob = glob
        self.node_list = node_list
        self.prescribes_dict = {}
        self.prescribed_freedoms = []
        # Format prescribes_dict = {freedomNo(world) : [values LC1, values LC2..])
        # vals are the magnitude of the prescribe -
        # all in world numbering 1, 2 , 3 etc,
        
    def modify(self, k_structure, p_structure):
#    def modify(self, k_structure, p_structure, node_list):
        '''
        Modify k_structure and p_structure for effects of prescribed displacements.
        Called by getKsPsUs.
        '''
        glob = self.glob
        if glob.noPrescribes == 0:
            return k_structure, p_structure  
            # Quit without modifying k_structure and p_structure
        fout  = glob.fout
        nn  = glob.no_freedoms
        big_no = 0
        # pick the larges diagonal term in the stiffness matrix
        for i in range(nn):
            if k_structure[i, i] > big_no:
                big_no = k_structure[i, i]
        big_no = big_no * 1e20
#        print 'd* big_no = ', big_no
        prescrdict = self.prescribes_dict   # alias
        print('prescrdict = ', prescrdict)
        freedomvec = list(prescrdict.keys())
        for i in range(len(freedomvec)):
#            freedom = self.prescribed_freedoms[i] - 1
            freedom = self.prescribed_freedoms[i]
#            print 'd* freedom = ', freedom
            k_structure[freedom, freedom] = big_no
            for ii in range(glob.no_load_cases):
                try:
#                    value = prescrdict[freedom + 1][ii]
                    value = prescrdict[freedom][ii]
                except KeyError:
                    cm_.printer(fout, 'num such key')
                    raise KeyError
                p_structure[freedom, ii] = value * big_no
        return k_structure, p_structure
        
    def read_prescribes(self):
        '''
        Create dictionary "prescribeDict" with key = freedom and
        value = list of values, no_loads (load cases) long. Called by main().
        '''
        glob = self.glob 
        fin = glob.fin
        fout = glob.fout
        printline = glob.qprint
        if self.glob.noPrescribes == 0:
            cm_.getTitle(fin, fout, printline, 'end')
            return
        else:
            dic = self.prescribes_dict
            prescribed_freedoms = list(dic.keys())  #  self.prescribed_freedoms            
            cm_.getTitle(fin, fout, printline, 'prescribes')
            for ii in range(glob.no_load_cases):
                for i in range(glob.noPrescribes):
                    lst = cm_.listRead(fin, fout, printline)
                    if lst[0] == 'b':
                        line = cm_.listProcess(fout, printline, lst, 'sisssfff')[1:]
                        #  b  3  n y n    0     -10e-3    0 
                        node = self.node_list.dic[line[0]]
                        # directory format:
                        # {nodeNo : [bt, ms, x, y,
                        #            flagX='f', flagY='f', flagXY='f']}
                        freedoms = [node[4], node[5], node[6]] # world units
                        use_flags = [line[1], line[2], line[3]]
                        values   = [line[4], line[5], line[6]]

                    elif lst[0] == 't':
                        line = cm_.listProcess(fout, printline, lst, 'sissff')[1:]
                        node = self.node_list.dic[line[0]]
                        # directory format:
                        # {nodeNo : [bt, ms, x, y, flagX='f', flagY='f']}
                        freedoms = [node[4], node[5]] # world units
                        use_flags = [line[1],    line[2]]
                        values   = [line[3],    line[4]]
                    else:
                        cm_.printer(fout, lst.__str__())
                        cm_.printer(fout, 'Found type = ' + lst[0])
                        cm_.abandon('Type must be either (b) or (t)')                       
                    if ii == 0:
                        for j in range(len(use_flags)):
                            if use_flags[j] == 'y':
                                dic[freedoms[j]] = [values[j]]
                                prescribed_freedoms.append(freedoms[j])
                    elif ii > 0:
                        for j in range(len(use_flags)):
                            if use_flags[j] == 'y':
                                dic[freedoms[j]].append(values[j])
            self.prescribed_freedoms = prescribed_freedoms  # world values
            self.prescribes_dict = dic  # world values
        cm_.getTitle(fin, fout, printline, 'end')
# End of Definition of object classes -------------------------------------

if __name__ == '__main__':
    print('hi from the joint_mod3.py.')
