#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
mainline3.py - a module for fm.py
'''
#
#        (c) Copyright 2009 Algis Kabaila.  All rights reserved.
#   
# This document is free; you can redistribute it and/or modify it under the terms
# of the GNU General Public License as published by the Free Software Foundation;
# either version 2 of the License, or (at your option) any later version.
#
# This document is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this document; if not, write to the Free Software Foundation, Inc.,
# 675 Mass Ave, Cambridge, MA 02139, USA.

import sys

from core_mods import core_mod3 as cm_
from core_mods import Matalg as _m

# Global functions and classes------------------------------------------------

DEBUG = False  #True # False

def d_print(line):
    '''Print string line on the console if debug is True.'''
    #    main_.d_print('\nLoad Matrix')
    if DEBUG:       
        print('*d* ', line)
    else:
        pass # do nothing.

def print_dictionary(qprint, dic):
    '''Neat printout of a dictionary.'''
    qprint('\ndic = \n')
    keys = list(dic.keys())
    keys.sort()
    for key in keys:
        qprint (key.__str__() + '   ' + dic[key].__str__() + '\n')  


class Structure_stiffness:
    '''Process all related to Structure Stiffness.'''
    def __init__(self, glob, props, load_matrix):
        '''pass all initial parameters to Structure Stiffness.'''
        self.glob = glob
        self.props = props
        self.load_matrix   = load_matrix
        no_freedoms = glob.no_freedoms
        k_structure = _m.Matrix(no_freedoms, no_freedoms)
        self.k_structure = k_structure
        
    def assemble_stiffness(self):
        '''
        Generate Structure Stiffness from element stiffnesses 
        and element location vectors.
        '''
        glob = self.glob
        props = self.props
        i_list = list(props.keys())        
        i_list.sort()
        for i in i_list:
            member = props[i]
            tag = member[0]
            d_print(tag)            
            if not(tag in ['t', 'b', 'u']):
                print('unrecognized member tag in')
                print('assemble_stiffness method of structureStiffness class')
                sys.exit(['tag not recognized'])
            if tag == 't':
                truss = glob.instances['truss']
                d_x, d_y, e_mod, area = member[2:6]
                loc_vect = member[6]
                k = truss.k_element(d_x, d_y, e_mod, area)
            elif tag == 'b':
                beam = glob.instances['beam']
                d_x, d_y, e_mod, area, i_x = member[2:7]
                loc_vect = member[7]
                k = beam.k_element(d_x, d_y, e_mod, area, i_x)
            elif tag == 'u':
                truss3dim = glob.instances['truss3dim']
                d_x, d_y, d_z, e_mod, area = member[2:7]
                loc_vect = member[7]
                k = truss3dim.k_element(d_x, d_y, d_z, e_mod, area)
            k_structure = self.k_structure
            k_structure = cm_.assemble(k_structure, k, loc_vect)
#        cm_.printer(fout, glob.qprint, '\nStructure Stiffness Matrix, k_structure.')
        print('\nStructure Stiffness Matrix, k_structure.')
        print(k_structure)
        return k_structure
# -------------------------------------------------------------------

    def find_el_disps(self, no_load_cases, loc_vector, u_s):
        '''Extract element displacements, u_e, from structure displacements u_s.
        '''
        size = len(loc_vector)
        u_e = _m.Matrix(size, no_load_cases)
        row = -1
        for var in loc_vector:
            i = abs(var) - 1 # Convert from world to program units.
            row += 1
            for j in range(no_load_cases):
                if i < 0:
                    u_e[row, j] = 0
                else:
                    u_e[row, j] = cm_.sgn(var) * u_s[i, j]
        return u_e

# -------------------------------------------------------------------

    def solve(self, mat_):
        '''
        Solve structure stiffness equations for structure displacements.
        Then extract the element displacements and calculate stresses.
        '''
        glob = self.glob
        fout = glob.fout
        qprint = glob.qprint
        load_matrix = self.load_matrix
        qprint('Load matrix.\n')
        qprint(load_matrix.__str__())
        props = self.props
        k_structure = self.k_structure
        u_s = k_structure ** load_matrix
        cm_.printer(fout, glob.qprint, 'Structure displacements (transposed matrix).')
        uss = u_s.mattranspose()
        print(uss)
        i_list = list(props.keys())
        cm_.printer(fout, qprint, "  Member stress resultants.")
        header = "  Member  LC    Axial        Shear        B.M. 1       B.M. 2\n"
        qprint(header)
        fout.write(header)
        i_list.sort()
        for i in i_list:
            member = props[i]
            tag = member[0]
            assert tag in ('t', 'b', 'u'), \
            'Unknown element tag in solve subroutine'
            if tag == 't':
                truss = glob.instances['truss']
                d_x, d_y, e_mod, area = member[2:6]
                loc_vect = member[6]
                stress_matrix = truss.stress_matrix(d_x, d_y, e_mod, area)
                u_e = self.find_el_disps(glob.no_load_cases, loc_vect, u_s)
                stress = stress_matrix * u_e
                for ii in range(len(stress[0])):                    
                    cm_.printer(self.glob.fout, qprint, "%5d %5d %12.5E " %\
                        (i, ii + 1, stress[0, ii]))
                cm_.printer(self.glob.fout, qprint, '  ')
            elif tag == 'b':
#    def beam_stress(self, e_mod, d_x, d_y, area, i_x, u_e, i):
#        '''Calculate and print beam stresses.'''
                beam = self.glob.instances['beam']
                d_x, d_y, e_mod, area, i_x = member[2:7]
                loc_vect = member[7]
                stress_matrix = beam.stress_matrix(d_x, d_y, e_mod, area, i_x)
                u_e = self.find_el_disps(glob.no_load_cases, loc_vect, u_s)
                stress = stress_matrix * u_e
                for ii in range(len(stress[0])):                    
                    cm_.printer(self.glob.fout, qprint, "%5d %5d %12.5E %12.5E %12.5E %12.5E" %\
                        (i, ii+1, stress[0,ii], stress[1,ii], stress[2,ii], stress[3,ii]))
                cm_.printer(self.glob.fout, qprint, '  ')
                
            elif tag == 'u':
                truss3dim = glob.instances['truss3dim']
                d_x, d_y, d_z, e_mod, area = member[2:7]
                loc_vect = member[7]
                stress_matrix = truss3dim.stress_matrix(d_x, d_y, d_z, e_mod, area)
                u_e = self.find_el_disps(glob.no_load_cases, loc_vect, u_s)
                stress = stress_matrix * u_e
                for ii in range(len(stress[0])):                    
                    cm_.printer(self.glob.fout, qprint, "%5d %5d %12.5E " %\
                        (i, ii + 1, stress[0, ii]))                       
        return u_s

class Global_variables:
    '''Read and store all important global variables. Expected to
    stay always in top level.'''
    def __init__(self, fin, fout, qprint, filename):
        self.fin  = fin
        self.fout = fout
        self.qprint = qprint
        self.filename = filename
        cm_.printer(fout, qprint, '\nSolution and debug echo of data.')
        cm_.getTitle(fin, fout, qprint, 'general')       
        # Read element types given in data and check their implementation
        mod_keys = cm_.listRead(fin, fout, qprint)
        # check if the specified element type has been implemented 
        modules = {}
        for var in mod_keys:
            if var == 'truss':
                from core_mods import truss_mod 
                modules['truss'] = truss_mod.Truss
            elif var == 'beam':
                from core_mods import beam_mod
                modules['beam'] = beam_mod.Beam
            elif var == 'truss3dim':
                from core_mods import truss3dim_mod
                modules['truss3dim'] = truss3dim_mod.Truss3dim
            else:
                qprint('unrecognised element type')
                raise My_exception
        self.modules = modules
        keys = list(modules.keys())
        for key in keys:
            qprint(modules[key].dic.__str__())
        self.instances = {}
#-----------------------------------------------------------------------------
        # Continue to process the elements
        lst = cm_.listRead(fin, fout, qprint)
        if len(lst) == 7:
            fmt = 'siiiiii'
        elif len(lst) == 8:
            fmt = 'siiiiiii'
        else:
            sys.exit(['wrong number of global variables'])
        lss  = cm_.listProcess(fout, qprint, lst, fmt)[1:]
        self.no_materials  = lss[0]
        self.no_sections   = lss[1]
        self.no_nodes      = lss[2]  
        self.no_members    = lss[3] 
        self.no_load_cases  = lss[4]
        self.noPrescribes = lss[5]
        self.no_element_types = 1
        if len(lss) > 6: 
            self.no_element_types = lss[6]
        # num of Loads in each load case.
        self.no_loads = []
        self.no_freedoms = None    # to be established by the program.
        lst = cm_.listRead(fin, fout, qprint)[1:]
        self.no_loads = [int(var) for var in lst]

    def __str__(self):
        s = "\nGlobal Data:\nNo of Materials = " + str(self.no_materials)
        s +=  "\nNo of Sections  = " + str(self.no_sections)
        s += "\nNo of Nodes     = " + str(self.no_nodes) + "\nNo of Members   = "
        s += str(self.no_members) + "\nNo of Load Cases= " + str(self.no_load_cases)
        s += "\nNo of Loads     = " + str(self.no_loads)
        s += "\nNo of prescribed displacements= " + str(self.noPrescribes)
        s += '\nNo of Element Types = ' + str(self.no_element_types)
        s += '\nNo of Freedoms = ' + str(self.no_freedoms)
        return s + '\nElement types: ' +  list(self.modules.keys()).__str__()
# -------------------------------------------------------------------           
    
if __name__ == '__main__':
    DEBUG = False #True # False
    print('dummy, to be deleted in the fullness of time.')


   
    
