#!/usr/bin/env python
''' truss_mod.py - module.'''

from math import sqrt

class Truss3dim:
    '''3 dimensional truss element properties for all data.
    '''
    dic = {'element_type' : 'truss3dim', 'tag' : 'u',  'material_items' : 1, 
    'section_items' : 1, 'coord_no' : 3, 'freedom_no' : 3, 'no_of_nodes' : 2,
    'nodal_load_no' : 3} 

    def __init__(self, mat_):
        self.mat_ = mat_

    def find_element_properties(self, line, node_list_dic,\
                                 material_list_dic, section_list_dic):
        '''line = member_input_line = member no, tag, end1 no, end2 no, 
        section no, material no. Expected to return 
        (tag, end1 X,Y,Z end2 X,Y,Z, section, E, location_vector). '''
        tag = line[0]
        if tag is not 'u':
            raise Exception, 'Can only handle truss3d member data.'
#        line = ['u', 2, 3, 1, 1]
#        line = member_input_line
        node1 = node_list_dic[line[1]]
        node2 = node_list_dic[line[2]]
        d_x = node2[3] - node1[3]
        d_y = node2[4] - node1[4]
        d_z = node2[5] - node1[5]
        loc_vect = node1[6:9] + node2[6:9]
        material_no = line[3]
        section_no = line[4]
        e_mod = material_list_dic[material_no]
        area  = section_list_dic[section_no]
        return d_x, d_y, d_z, e_mod, area, loc_vect
    
    def bmat_kbar(self, d_x, d_y, d_z, e_mod, area):
        "Compute matrix B and member stiffness kbar for truss bar."
        mat_ = self.mat_
        b_mat = mat_.zerosM(6, 1)
        length = sqrt(d_x * d_x + d_y * d_y + d_z * d_z)
        cosx = d_x/length
        cosy = d_y/length
        cosz = d_z/length
        b_mat[0, 0] = -cosx
        b_mat[1, 0] = -cosy
        b_mat[2, 0] = -cosz
        b_mat[3, 0] =  cosx
        b_mat[4, 0] =  cosy
        b_mat[5, 0] =  cosz
        kbar = (e_mod * area) / length
        return b_mat, kbar    

    def k_element(self, d_x, d_y, d_z, e_mod, area):
        'Compute k_mat for truss bar.'
        mat_ = self.mat_
        b_mat, kbar = self.bmat_kbar(d_x, d_y, d_z,  e_mod, area)
        b_transposed = mat_.transposeM(b_mat)
        k_mat = kbar * mat_.multM(b_mat, b_transposed)
        return  k_mat

    def stress_matrix(self, d_x, d_y, d_z, e_mod, area):
        '''Returns truss truss bar's stress matrix.'''
        mat_ = self.mat_
        b_mat, kbar = self.bmat_kbar(d_x, d_y, d_z, e_mod, area)
        b_transposed = mat_.transposeM(b_mat)
        # kbar is a scalar for truss member
        stress_mat = kbar * b_transposed
        return stress_mat

# end of 3d truss functions -----------------------------------------

if __name__ == '__main__':
# ---------------------------------------------
    ok = True
    try:
        import matrixM
    except ImportError:
        ok = False
    print
# for the import to work, the matrixM module needs to be on Python path
# and numpy is installed.
    print
    keys = Truss3dim.dic.keys()
    print 'Dictionary "Truss.dic" ='
    print Truss3dim.dic
    print 'Is import of matrixM successful? Ans: ', ok
# The following is only attempted if the matrixM is imported.
    if ok:
        mat_ = matrixM.Matrix('numpy')
        truss3dim = Truss3dim(mat_)
        keys = Truss3dim.dic.keys()
        print 'dictionary "dic" keys ='
        print keys
        print 'Dictionary "dic" ='
        print Truss3dim.dic
        mat_ = matrixM.Matrix('numpy')
    
        node_list_dic = {1: ['u', 'm', '1', 0.0, 0.0, 0.0, 0, 0, 0], 
                         2: ['u', 'm', '2', 0.0, 6.0, 0.0, 0, 0, 0], 
                         3: ['u', 'm', '3', 8.0, 6.0, 0.0, 0, 0, 0],
                         4: ['u', 'm', '4', 8.0, 0.0, 0.0, 0, 0, 0], 
                         5: ['u', 'm', '5', 4.0, 3.0, 5.0, 1, 2, 3] }
    
        member_list_dic = {1: ['u', 1, 5, 1, 1], 
                           2: ['u', 2, 5, 1, 2], 
                           3: ['u', 3, 5, 1, 3], 
                           4: ['u', 4, 5, 1, 4]}
        material_list_dic =   {1: 200000.0}
        section_list_dic =    {1: 4.0, 2: 3.0, 3: 2.0, 4: 1.0}

# Take member no 3, for example
        line = member_list_dic[3]
        print line
        properties = truss3dim.find_element_properties(line, node_list_dic,
                                                        material_list_dic, section_list_dic)
        print 'properties = (d_x, d_y, d_z, e_mod, area [loc_vect]) = \n', properties
# Stiffness matrix
        d_x, d_y, d_z, e_mod, area,  loc_vect = properties
        k_el = truss3dim.k_element(d_x, d_y, d_z, e_mod, area)
        print '\nelement stiffness matrix, k_el = '
        mat_.printM(k_el)
#  Stress matrix
        stress_mat = truss3dim.stress_matrix(d_x, d_y, d_z, e_mod, area)
        print 'stress matrix '
        mat_.printM(stress_mat)
